# TODO: update any doc strings as needed in here
# TODO: ensure that 'Default' setting takes None or 0 as a node number (None is all-nodes, 0 is UNKNOWN/UNINTERPRETED). Think I did this for None.

from abc import ABC
from collections import defaultdict
from pathlib import Path
from typing import List, Union
import importlib
import pandas as pd

import emod_api
from emod_api.schema_to_class import ReadOnlyDict

import emodpy_hiv.campaign.cascade_of_care as coc
from emodpy.emod_task import EMODTask
from emodpy_hiv.campaign.common import TargetGender
from emodpy_hiv.demographics.hiv_demographics import HIVDemographics
from emodpy_hiv.parameterized_call import ParameterizedCall
from emodpy_hiv.reporters.reporters import Reporters


def get_country_class(country_class_name: str):
    """
    Import the countries module and return the class of the country.

    Args:
        country_class_name:
            The name of the country to get the class for.  The name is expected to be exactly like
            like it is in its module.  Countries with more than one word are expected to have an
            underscore ('_') between the words.

            Examples of naming convention:
            - Zambia should be located in the module emodpy_hiv.countries.zambia.zambia
            - South Africa should have a class called South_Africa and be in a module called
            emodpy_hiv.countries.south_africa.south_africa
    """
    module_name = f"emodpy_hiv.countries.{country_class_name.lower()}.{country_class_name.lower()}"
    module_name = module_name.replace(" ", "_")
    country_module = importlib.import_module(module_name)
    country_class = getattr(country_module, country_class_name)
    return country_class


class Country(ABC):
    """
    Country is an "abstract" base class for HIV models of different countries or regions.
    It is intended that a person wanting to develop a new model of a country will either
    subclass Country (or another Country model) and change the things that are different.
    One could use this to ensure that things that are supposed to be the same across all
    of the country models are the same.

    WARNING: Country models are used as the class and not an instance object.  That is,
    one does not create an object like:
    * my_country = MyCountry()
    Instead, they call the classmethods directly from the class like:
    * demographics = MyCountry.build_demographics()
    If someone attempts to create an instance, we want it to throw an exception so people
    don't use it in a way that is not intended.

    WARNING_2: As of Python 3.13, you are not supposed to combine different properties
    with the classmethod property. It should be used by itself.  This is way we are using
    class variables instead of properties.  When we do class inheritance, each subclass
    gets its own instance of the class variables.  For example, if Zambia and Kenya are
    both subclasss of Country, they will have their own instance of country_name even
    though it is declared in Country.  This allows us to set country_name for Zambia
    without it impacting Kenya.

    NOTE: We are using classes instead of modules because we can still get inheritance
    like you'd expect.  For example, with classes if BaseCountry has three methods:
    func_A(), func_B(), and func_C().  Assume func_C() calls both func_A() and func_B().
    Let's alo assume that we create Zambia as a subclass of BaseCountry and have it
    override func_B().  With classes, when func_C()) is called for Zambia, it will use
    the overridden func_B().  If we tried to do the same thing with modules, func_C()
    would use the func_B() defined in BaseCountry.

    We use Country as a class and this idea of "parameterized calls" so that the country
    can play well with emod_workflow.  The "parameterized calls" need the functions to be
    static and these calls allow us to give a unique name to each parameter.   Unique
    parameter names allow the user to specifically refer to a parameter even when there
    are numerous instances of the object that has the parameter.

    """

    # ------------------------------------------------------------------------------------
    # --- NOTE: I tried using the "Attributes:" stuff that Google says to use, but I kept
    # --- getting this #duplicate object description ... use :noindex: for one of them"
    # --- error message from ReadTheDocs.
    # ------------------------------------------------------------------------------------

    # String: The name of the country model
    country_name = None

    # Float: The calendar year to start the simulation.  This is frequently
    # set to 1960.5 so that there is time to burn-in the population and relationships.
    base_year = 1960.5

    def __init__(self):
        msg =  "You cannot create an instance of a Country.\n" # noqa: E222
        msg += "You must use the class as the object."
        raise NotImplementedError(msg)

    @staticmethod
    def generate_label(relationship_type: str = None,
                       risk_group: str = None,
                       node_ids: Union[List[int], None] = None,
                       node_name: str = None) -> str:
        if node_name is not None:
            label_items = [str(item) for item in [relationship_type, risk_group, node_name] if item is not None]
        else:
            node_ids = [] if node_ids is None else sorted(node_ids)
            label_items = [str(item) for item in [relationship_type, risk_group, *node_ids, node_name] if item is not None]
        label = '-'.join(label_items)
        return label

    @classmethod
    def get_config_parameterized_calls(cls, config: ReadOnlyDict) -> List[ParameterizedCall]:
        """
        Return a list of ParameterizedCall objects that will complete the configuration of an
        EMOD config object.

        Args
            config: The object to be updated when using instance methods with ParameterizedCall.

        Returns:
            List of ParameterizedCall objects
        """
        parameterized_calls = []

        def set_run_number(config, Run_Number: int = None):
            if Run_Number is not None:
                config.parameters.Run_Number = Run_Number

        pc = ParameterizedCall(func=set_run_number, hyperparameters={'Run_Number': None})
        parameterized_calls.append(pc)
        return parameterized_calls

    @classmethod
    def get_demographics_parameterized_calls(cls, demographics: HIVDemographics) -> List[ParameterizedCall]:
        """
        Return a list of ParameterizedCall objects that will complete the configuration of an
        EMOD demographics object.

        Args
            demographics: If not None, this will be the object that is updated

        Returns:
            List of ParameterizedCall objects
        """
        return []

    @classmethod
    def get_campaign_parameterized_calls(cls, campaign: emod_api.campaign) -> List[ParameterizedCall]:
        """
        Return a list of ParameterizedCall objects that will complete the configuration of an
        EMOD campaign object.

        Args
            campaign: If not None, this will be the object that is updated

        Returns:
            List of ParameterizedCall objects
        """
        return []

    @classmethod
    def initialize_config(cls, schema_path: Union[str, Path]) -> ReadOnlyDict:
        """
        Override this in the subclass to set the default values for the config object as needed.
        """
        config = EMODTask.build_default_config(schema_path=schema_path)
        return config

    @classmethod
    def initialize_demographics(cls) -> HIVDemographics:
        """
        Create the initial demographics object that will be modified by the ParameterizedCall
        objects from **get_demographics_parameterized_calls()**
        """
        demographics = HIVDemographics.from_template_node(pop=10000,
                                                          default_society_template="PFA-Southern-Africa")

        # TODO:  How to load default node attributes?

        return demographics

    @classmethod
    def initialize_campaign(cls, schema_path: Union[str, Path]) -> emod_api.campaign:
        """
        Create the initial campaign object that will be modified by the ParameterizedCall
        objects from **get_campaign_parameterized_calls()**
        """
        campaign = EMODTask.build_default_campaign(schema_path=schema_path)
        campaign.base_year = cls.base_year
        return campaign

    #
    # Convenience functions for non-EMOD-workflow users below
    #

    @classmethod
    def build_reports(cls, reporters: Reporters) -> Reporters:
        """
        Function that creates reports to be used in the experiment. The function must have Reporters
        object as the parameter and return that object. It is assumed that all the reporters come from the
        reporters that are part of EMOD main code. EMOD also supports reporters as custom plug-in dlls,
        however, not through the current emodpy system.

        Args:
            reporters: The Reporters object to add reports to

        Returns:
            Reporters object
        """
        return reporters

    @classmethod
    def _execute_parameterized_calls_on(cls, obj, parameterized_calls: List[ParameterizedCall]):
        for parameterized_call in parameterized_calls:
            # Now modify the provided object (obj). obj is passed as context to each successive ParameterizedCall.
            prepared_call = parameterized_call.prepare_call()
            prepared_call(obj)

    @classmethod
    def build_config(cls, config: ReadOnlyDict) -> ReadOnlyDict:
        """
        A function that creates a config object that can be used with
        EMODTask.from_defaults().

        Args:
            config(ReadOnlyDict): the config object that will be modified. It has the following structure:
                config.parameters.<parameter_name> , with <parameter_name> being an attribute.
        Returns:
            a config object
        """
        config = cls.initialize_config(schema_path=config["schema_path"])
        cls._execute_parameterized_calls_on(obj=config, parameterized_calls=cls.get_config_parameterized_calls(config=config))
        return config

    @classmethod
    def build_demographics(cls) -> HIVDemographics:
        """
        A function that configures the demographics for EMOD and can be used with
        EMODTask.from_defaults().

        Returns:
            a demographics object
        """
        demographics = cls.initialize_demographics()
        cls._execute_parameterized_calls_on(obj=demographics,
                                            parameterized_calls=cls.get_demographics_parameterized_calls(demographics=demographics))
        return demographics

    @classmethod
    def build_campaign(cls, campaign: emod_api.campaign) -> emod_api.campaign:
        """
        A method to pass into EMODTask to build campaign events for Emod.

        Args:
            campaign(emod_api.campaign): The emod_api campaign object to be modified.
        Returns:
            a campaign object
        """
        campaign = cls.initialize_campaign(schema_path=campaign.schema_path)
        calls = cls.get_campaign_parameterized_calls(campaign=campaign)
        cls._execute_parameterized_calls_on(obj=campaign, parameterized_calls=calls)
        return campaign

    #
    # common functions used for building campaign ParameterizedCalls below
    #

    @staticmethod
    def load_nchooser_distribution_data(file_path: str) -> dict:
        """
        Load target distribution data into a dataframe format that works for Nchooser and group by node_id.
        Args:
            file_path: path to the csv file that contains the data

        Returns:
            dict: a dictionary that has node_id as the key and a dataframe as the value. The dataframe contains the
            distribution data for Nchooser.

        """
        data = pd.read_csv(file_path)
        data_dict = defaultdict(pd.DataFrame)
        for node_id in pd.unique(data['node_set']):
            node_data = data[data['node_set'] == node_id]
            node_dict = {'year': [], 'min_age': [], 'max_age': [], 'n_circumcisions': []}
            for year in pd.unique(node_data['year']):
                year_data = node_data[node_data['year'] == year]
                for _, row in year_data.iterrows():
                    node_dict['year'].append(year)
                    node_dict['min_age'].append(float(row['age_bin'].split(':')[0].replace('[', '')))
                    node_dict['max_age'].append(float(row['age_bin'].split(':')[1].replace(')', '')) - 0.00001)
                    node_dict['n_circumcisions'].append(int(row['n_circumcisions']))
            data_dict[int(node_id)] = pd.DataFrame(node_dict)
        return data_dict

    @classmethod
    def add_historical_vmmc_nchooser(cls,
                                     campaign: emod_api.campaign,
                                     historical_vmmc_data_filepath: str,
                                     historical_vmmc_reduced_acquire: float = 0.6):
        # Load historical vmmc data into data frame
        historical_vmmc_data = cls.load_nchooser_distribution_data(historical_vmmc_data_filepath)
        for node_id in historical_vmmc_data:
            coc.add_historical_vmmc_nchooser(campaign=campaign,
                                             historical_vmmc_distributions_by_time=historical_vmmc_data[node_id],
                                             historical_vmmc_reduced_acquire=historical_vmmc_reduced_acquire,
                                             historical_vmmc_property_restrictions=None,
                                             historical_vmmc_node_ids=[node_id],
                                             has_intervention_name_exclusion=coc.ANY_MC,
                                             event_name=f"NChooser to produce specified number of program VMMCs in node"
                                                        f" {node_id}, {cls.country_name}")

    @classmethod
    def add_traditional_male_circumcision(cls,
                                          campaign: emod_api.campaign,
                                          traditional_male_circumcision_coverage: float = 0.5,  # TODO: what is a meaningful default?
                                          traditional_male_circumcision_reduced_acquire: float = 0.6,
                                          node_ids: Union[List[int], None] = None):
        randomchoice_start_year = 1975
        traditional_male_circumcision_start_year = 1961
        coc.add_traditional_male_circumcision(campaign=campaign,
                                              traditional_male_circumcision_start_year=traditional_male_circumcision_start_year,
                                              randomchoice_start_year=randomchoice_start_year,
                                              traditional_male_circumcision_coverage=traditional_male_circumcision_coverage,
                                              traditional_male_circumcision_reduced_acquire=traditional_male_circumcision_reduced_acquire,
                                              traditional_male_circumcision_node_ids=node_ids)

    @classmethod
    def add_vmmc_reference_tracking(cls,
                                    campaign: emod_api.campaign,
                                    tracking_vmmc_reduced_acquire: float = 0.6):
        vmmc_time_value_map = {
            "Times": [2015.9999, 2016, 2021],
            "Values": [0, 0.54, 0.9]  # Zambia values
        }
        coc.add_vmmc_reference_tracking(campaign=campaign,
                                        vmmc_time_value_map=vmmc_time_value_map,
                                        vmmc_reduced_acquire=tracking_vmmc_reduced_acquire,
                                        vmmc_target_min_age=15,
                                        vmmc_target_max_age=15.09,
                                        vmmc_start_year=2015,
                                        vmmc_node_ids=None,  # all nodes
                                        update_period=30.4166666666667)

    @classmethod
    def seed_infections(cls,
                        campaign: emod_api.campaign,
                        node_ids: Union[List[int], None] = None,
                        seeding_start_year: float = 1982,
                        seeding_coverage: float = 0.2,
                        seeding_target_min_age: float = 0,
                        seeding_target_max_age: float = 200,
                        seeding_target_gender: TargetGender = TargetGender.ALL):
        # add outbreak individual interventions with coverage 0.1 to all nodes with Risk:HIGH
        coc.seed_infections(campaign=campaign,
                            seeding_node_ids=node_ids,
                            seeding_start_year=seeding_start_year,
                            seeding_coverage=seeding_coverage,
                            seeding_target_min_age=seeding_target_min_age,
                            seeding_target_max_age=seeding_target_max_age,
                            seeding_target_gender=seeding_target_gender,
                            seeding_target_property_restrictions=["Risk:HIGH"])

    @classmethod
    def add_csw(cls,
                campaign: emod_api.campaign,
                csw_male_uptake_coverage: float = 0.1,
                csw_female_uptake_coverage: float = 0.1,  # TODO: meaningful default values?
                node_ids: Union[List[int], None] = None):
        coc.add_csw(campaign=campaign,
                    node_ids=node_ids,
                    male_uptake_coverage=csw_male_uptake_coverage,
                    female_uptake_coverage=csw_female_uptake_coverage)

    @classmethod
    def add_post_debut_coinfection(cls,
                                   campaign: emod_api.campaign,
                                   coinfection_coverage_HIGH: float = 0.3,
                                   coinfection_coverage_MEDIUM: float = 0.3,
                                   coinfection_coverage_LOW: float = 0.1,
                                   node_ids: Union[List[int], None] = None):
        coinfection_risk_to_coverage = {
            'HIGH': coinfection_coverage_HIGH,
            'MEDIUM': coinfection_coverage_MEDIUM,
            'LOW': coinfection_coverage_LOW
        }
        for risk, coverage in coinfection_risk_to_coverage.items():
            coc.add_post_debut_coinfection(campaign=campaign,
                                           coinfection_coverage=coverage,
                                           coinfection_gender=TargetGender.ALL,
                                           coinfection_IP=f'Risk:{risk}',
                                           coinfection_node_ids=node_ids)

    @classmethod
    def add_pmtct(cls, campaign: emod_api.campaign, node_ids: Union[List[int], None] = None):
        cls.add_state_TestingOnANC(campaign, node_ids=node_ids)
        # Testing of children at 6 weeks of age
        cls.add_state_TestingOnChild6w(campaign, node_ids=node_ids)

    @classmethod
    def add_health_care_testing(cls,
                                campaign: emod_api.campaign,
                                hct_delay_to_next_test: List[float],
                                hct_delay_to_next_test_node_ids: List[List[int]],
                                hct_delay_to_next_test_node_names: List[str],
                                hct_start_year: float = 1990,
                                node_ids: Union[List[int], None] = None):

        cls.add_state_HCTUptakeAtDebut(campaign=campaign, start_year=hct_start_year, node_ids=node_ids)

        cls.add_state_HCTUptakePostDebut(campaign=campaign, start_year=hct_start_year, node_ids=node_ids)

        cls.add_state_HCTTestingLoop(campaign=campaign,
                                     hct_delay_to_next_test=hct_delay_to_next_test,
                                     hct_delay_to_next_test_node_ids=hct_delay_to_next_test_node_ids,
                                     hct_delay_to_next_test_node_names=hct_delay_to_next_test_node_names,
                                     start_year=hct_start_year,
                                     node_ids=node_ids)

    @classmethod
    def add_ART_cascade(cls,
                        campaign: emod_api.campaign,
                        art_cascade_start_year: float = 1990,
                        cd4_retention_rate: float = 1,
                        linking_to_pre_art_sigmoid_min: float = 0.7572242198,
                        linking_to_pre_art_sigmoid_max: float = 0.9591484679,
                        linking_to_pre_art_sigmoid_midyear: float = 2006.8336631523,
                        linking_to_pre_art_sigmoid_rate: float = 1.0,
                        pre_staging_retention: float = 0.85,
                        pre_art_retention: float = 0.75,
                        linking_to_art_sigmoid_min: float = 0.0,
                        linking_to_art_sigmoid_max: float = 0.8507390283,
                        linking_to_art_sigmoid_midyear: float = 1997.4462231708,
                        linking_to_art_sigmoid_rate: float = 1.0,
                        art_reenrollment_willingness: float = 0.9,
                        immediate_art_rate: float = 0.1,
                        node_ids: Union[List[int], None] = None):

        cls.add_state_TestingOnSymptomatic(campaign=campaign, start_year=art_cascade_start_year, node_ids=node_ids)

        #
        # ---- BEGIN ART STAGING SECTION ----
        #
        cls.add_state_ARTStagingDiagnosticTest(campaign=campaign, start_year=art_cascade_start_year, node_ids=node_ids)

        cls.add_state_ARTStaging(campaign=campaign, start_year=art_cascade_start_year,
                                 cd4_retention_rate=cd4_retention_rate,
                                 pre_staging_retention=pre_staging_retention,
                                 node_ids=node_ids)
        #
        # ---- BEGIN PRE-ART ----
        #
        # chance of linking to pre-ART
        cls.add_state_LinkingToPreART(campaign=campaign,
                                      start_year=art_cascade_start_year,
                                      sigmoid_min=linking_to_pre_art_sigmoid_min,
                                      sigmoid_max=linking_to_pre_art_sigmoid_max,
                                      sigmoid_midyear=linking_to_pre_art_sigmoid_midyear,
                                      sigmoid_rate=linking_to_pre_art_sigmoid_rate,
                                      node_ids=node_ids)

        # ensuring each agent continues this cascade once per timestep
        cls.add_state_OnPreART(campaign=campaign, start_year=art_cascade_start_year, pre_art_retention=pre_art_retention,
                               node_ids=node_ids)
        #
        # ---- BEGIN ART LINKING ----
        #
        cls.add_state_LinkingToART(campaign=campaign,
                                   start_year=art_cascade_start_year,
                                   sigmoid_min=linking_to_art_sigmoid_min,
                                   sigmoid_max=linking_to_art_sigmoid_max,
                                   sigmoid_midyear=linking_to_art_sigmoid_midyear,
                                   sigmoid_rate=linking_to_art_sigmoid_rate,
                                   node_ids=node_ids)

        # decide to initiate ART now or later
        cls.add_state_OnART(campaign=campaign, start_year=art_cascade_start_year,
                            art_reenrollment_willingness=art_reenrollment_willingness,
                            immediate_art_rate=immediate_art_rate,
                            node_ids=node_ids)
        #
        cls.add_state_LostForever(campaign=campaign, start_year=art_cascade_start_year, node_ids=node_ids)

    @classmethod
    def add_state_TestingOnANC(cls, campaign: emod_api.campaign, node_ids: Union[List[int], None] = None):
        start_year = 1990
        coverage = 1.0
        sigmoid_min = 0
        sigmoid_max = 0.975
        sigmoid_midyear = 2005.87
        sigmoid_rate = 0.7136
        link_to_ART_rate = 0.8
        treatment_a_efficacy = 0.9
        treatment_b_efficacy = 0.96667
        sdNVP_efficacy = 0.66
        disqualifying_properties = [coc.CascadeState.LOST_FOREVER,
                                    coc.CascadeState.ON_ART,
                                    coc.CascadeState.LINKING_TO_ART,
                                    coc.CascadeState.ON_PRE_ART,
                                    coc.CascadeState.LINKING_TO_PRE_ART,
                                    coc.CascadeState.ART_STAGING,
                                    coc.CascadeState.TESTING_ON_SYMPTOMATIC]
        property_restrictions = 'Accessibility:Yes'

        coc.add_state_TestingOnANC(campaign=campaign,
                                   disqualifying_properties=disqualifying_properties,
                                   coverage=coverage,
                                   link_to_ART_rate=link_to_ART_rate,
                                   node_ids=node_ids,
                                   sigmoid_min=sigmoid_min,
                                   sigmoid_max=sigmoid_max,
                                   sigmoid_midyear=sigmoid_midyear,
                                   sigmoid_rate=sigmoid_rate,
                                   treatment_a_efficacy=treatment_a_efficacy,
                                   treatment_b_efficacy=treatment_b_efficacy,
                                   sdNVP_efficacy=sdNVP_efficacy,
                                   start_year=start_year,
                                   property_restrictions=property_restrictions)

    @classmethod
    def add_state_TestingOnChild6w(cls, campaign: emod_api.campaign, node_ids: Union[List[int], None] = None):
        child_testing_start_year = 2004
        child_testing_time_value_map = {"Times": [2004, 2005, 2006, 2008, 2009],  # TODO: move to Zambia?
                                        "Values": [0, 0.03, 0.1, 0.2, 0.3365]}
        disqualifying_properties = [coc.CascadeState.LOST_FOREVER,
                                    coc.CascadeState.ON_ART,
                                    coc.CascadeState.LINKING_TO_ART,
                                    coc.CascadeState.ON_PRE_ART,
                                    coc.CascadeState.LINKING_TO_PRE_ART,
                                    coc.CascadeState.ART_STAGING,
                                    coc.CascadeState.TESTING_ON_SYMPTOMATIC]
        property_restrictions = 'Accessibility:Yes'
        coc.add_state_TestingOnChild6w(campaign=campaign,
                                       disqualifying_properties=disqualifying_properties,
                                       time_value_map=child_testing_time_value_map,
                                       node_ids=node_ids,
                                       property_restrictions=property_restrictions,
                                       start_year=child_testing_start_year)

    @classmethod
    def add_state_HCTUptakeAtDebut(cls,
                                   campaign: emod_api.campaign,
                                   start_year: float,
                                   node_ids: Union[List[int], None] = None):
        disqualifying_properties = [coc.CascadeState.LOST_FOREVER,
                                    coc.CascadeState.ON_ART,
                                    coc.CascadeState.LINKING_TO_ART,
                                    coc.CascadeState.ON_PRE_ART,
                                    coc.CascadeState.LINKING_TO_PRE_ART,
                                    coc.CascadeState.ART_STAGING]
        coc.add_state_HCTUptakeAtDebut(campaign=campaign,
                                       disqualifying_properties=disqualifying_properties,
                                       node_ids=node_ids,
                                       start_year=start_year)

    @classmethod
    def add_state_HCTUptakePostDebut(cls,
                                     campaign: emod_api.campaign,
                                     start_year: float,
                                     node_ids: Union[List[int], None] = None):
        disqualifying_properties = [coc.CascadeState.LOST_FOREVER,
                                    coc.CascadeState.ON_ART,
                                    coc.CascadeState.LINKING_TO_ART,
                                    coc.CascadeState.ON_PRE_ART,
                                    coc.CascadeState.LINKING_TO_PRE_ART,
                                    coc.CascadeState.ART_STAGING]
        coc.add_state_HCTUptakePostDebut(campaign=campaign,
                                         disqualifying_properties=disqualifying_properties,
                                         node_ids=node_ids,
                                         hct_reentry_rate=1,
                                         start_year=start_year,
                                         tvmap_test_for_enter_HCT_testing_loop=coc.all_negative_time_value_map)

    @classmethod
    def add_state_HCTTestingLoop(cls, campaign: emod_api.campaign,
                                 start_year: float,
                                 hct_delay_to_next_test: List[float],
                                 hct_delay_to_next_test_node_ids: List[List[int]],
                                 hct_delay_to_next_test_node_names: List[str],
                                 node_ids: Union[List[int], None] = None):
        disqualifying_properties = [coc.CascadeState.LOST_FOREVER,
                                    coc.CascadeState.ON_ART,
                                    coc.CascadeState.LINKING_TO_ART,
                                    coc.CascadeState.ON_PRE_ART,
                                    coc.CascadeState.LINKING_TO_PRE_ART,
                                    coc.CascadeState.ART_STAGING]
        coc.add_state_HCTTestingLoop(campaign=campaign,
                                     disqualifying_properties=disqualifying_properties,
                                     start_year=start_year,
                                     hct_retention_rate=0.95,
                                     tvmap_consider_immediate_ART=coc.all_negative_time_value_map,
                                     hct_delay_to_next_test=hct_delay_to_next_test,
                                     hct_delay_to_next_test_node_ids=hct_delay_to_next_test_node_ids,
                                     hct_delay_to_next_test_node_names=hct_delay_to_next_test_node_names,
                                     node_ids=node_ids)

    @classmethod
    def add_state_TestingOnSymptomatic(cls,
                                       campaign: emod_api.campaign,
                                       start_year: float,
                                       node_ids: Union[List[int], None] = None):
        disqualifying_properties = [coc.CascadeState.LOST_FOREVER,
                                    coc.CascadeState.ON_ART,
                                    coc.CascadeState.LINKING_TO_ART,
                                    coc.CascadeState.ON_PRE_ART,
                                    coc.CascadeState.LINKING_TO_PRE_ART,
                                    coc.CascadeState.ART_STAGING]
        coc.add_state_TestingOnSymptomatic(campaign=campaign,
                                           disqualifying_properties=disqualifying_properties,
                                           start_year=start_year,
                                           node_ids=node_ids,
                                           tvmap_increased_symptomatic_presentation=coc.all_negative_time_value_map)

    @classmethod
    def add_state_ARTStagingDiagnosticTest(cls,
                                           campaign: emod_api.campaign,
                                           start_year: float,
                                           node_ids: Union[List[int], None] = None):
        disqualifying_properties = [coc.CascadeState.LOST_FOREVER,
                                    coc.CascadeState.ON_ART,
                                    coc.CascadeState.LINKING_TO_ART,
                                    coc.CascadeState.ON_PRE_ART,
                                    coc.CascadeState.LINKING_TO_PRE_ART]
        coc.add_state_ARTStagingDiagnosticTest(campaign=campaign,
                                               disqualifying_properties=disqualifying_properties,
                                               start_year=start_year,
                                               node_ids=node_ids)

    @classmethod
    def add_state_ARTStaging(cls,
                             campaign: emod_api.campaign,
                             start_year: float,
                             cd4_retention_rate: float = 1,
                             pre_staging_retention: float = 0.85,
                             node_ids: Union[List[int], None] = None):
        disqualifying_properties = [coc.CascadeState.LOST_FOREVER,
                                    coc.CascadeState.ON_ART,
                                    coc.CascadeState.LINKING_TO_ART,
                                    coc.CascadeState.ON_PRE_ART,
                                    coc.CascadeState.LINKING_TO_PRE_ART]
        coc.add_state_ARTStaging(campaign=campaign,
                                 disqualifying_properties=disqualifying_properties,
                                 start_year=start_year,
                                 node_ids=node_ids,
                                 pre_staging_retention=pre_staging_retention,
                                 cd4_retention_rate=cd4_retention_rate)

    @classmethod
    def add_state_LinkingToPreART(cls,
                                  campaign: emod_api.campaign,
                                  start_year: float,
                                  sigmoid_min: float = 0.7572242198,
                                  sigmoid_max: float = 0.9591484679,
                                  sigmoid_midyear: float = 2006.8336631523,
                                  sigmoid_rate: float = 1.0,
                                  node_ids: Union[List[int], None] = None):
        disqualifying_properties_pre_art_linking = [coc.CascadeState.LOST_FOREVER,
                                                    coc.CascadeState.ON_ART,
                                                    coc.CascadeState.LINKING_TO_ART,
                                                    coc.CascadeState.ON_PRE_ART]
        coc.add_state_LinkingToPreART(campaign=campaign,
                                      disqualifying_properties=disqualifying_properties_pre_art_linking,
                                      start_year=start_year,
                                      sigmoid_min=sigmoid_min,
                                      sigmoid_max=sigmoid_max,
                                      sigmoid_midyear=sigmoid_midyear,
                                      sigmoid_rate=sigmoid_rate,
                                      node_ids=node_ids)

    @classmethod
    def add_state_OnPreART(cls,
                           campaign: emod_api.campaign,
                           start_year: float,
                           pre_art_retention: float = 0.75,
                           node_ids: Union[List[int], None] = None):
        disqualifying_properties_pre_art = [coc.CascadeState.LOST_FOREVER,
                                            coc.CascadeState.ON_ART,
                                            coc.CascadeState.LINKING_TO_ART]
        coc.add_state_OnPreART(campaign=campaign,
                               node_ids=node_ids,
                               pre_art_retention=pre_art_retention,
                               disqualifying_properties=disqualifying_properties_pre_art,
                               start_year=start_year)

    @classmethod
    def add_state_LinkingToART(cls,
                               campaign: emod_api.campaign,
                               start_year: float,
                               sigmoid_min: float = 0.0,
                               sigmoid_max: float = 0.8507390283,
                               sigmoid_midyear: float = 1997.4462231708,
                               sigmoid_rate: float = 1.0,
                               node_ids: Union[List[int], None] = None):
        disqualifying_properties = [coc.CascadeState.LOST_FOREVER, coc.CascadeState.ON_ART]
        coc.add_state_LinkingToART(campaign=campaign,
                                   node_ids=node_ids,
                                   disqualifying_properties=disqualifying_properties,
                                   start_year=start_year,
                                   sigmoid_min=sigmoid_min,
                                   sigmoid_max=sigmoid_max,
                                   sigmoid_midyear=sigmoid_midyear,
                                   sigmoid_rate=sigmoid_rate)

    @classmethod
    def add_state_OnART(cls,
                        campaign: emod_api.campaign,
                        start_year: float,
                        art_reenrollment_willingness: float = 0.9,
                        immediate_art_rate: float = 0.1,
                        node_ids: Union[List[int], None] = None):
        tvmap_immediate_ART_restart = coc.all_negative_time_value_map
        tvmap_reconsider_lost_forever = coc.all_negative_time_value_map
        disqualifying_properties = [coc.CascadeState.LOST_FOREVER]
        coc.add_state_OnART(campaign=campaign,
                            art_reenrollment_willingness=art_reenrollment_willingness,
                            immediate_art_rate=immediate_art_rate,
                            node_ids=node_ids,
                            disqualifying_properties=disqualifying_properties,
                            start_year=start_year,
                            tvmap_immediate_ART_restart=tvmap_immediate_ART_restart,
                            tvmap_reconsider_lost_forever=tvmap_reconsider_lost_forever)

    @classmethod
    def add_state_LostForever(cls,
                              campaign: emod_api.campaign,
                              start_year: float,
                              node_ids: Union[List[int], None] = None):
        coc.add_state_LostForever(campaign=campaign,
                                  node_ids=node_ids,
                                  start_year=start_year)



from emodpy.utils.emod_enum import StrEnum, DistributionType, NodeSelectionType, VaccineType, SensitivityType, \
    EventOrConfig, SettingType


class PrioritizePartnersBy(StrEnum):
    NO_PRIORITIZATION = 'NO_PRIORITIZATION'
    CHOSEN_AT_RANDOM = 'CHOSEN_AT_RANDOM'
    LONGER_TIME_IN_RELATIONSHIP = 'LONGER_TIME_IN_RELATIONSHIP'
    SHORTER_TIME_IN_RELATIONSHIP = 'SHORTER_TIME_IN_RELATIONSHIP'
    OLDER_AGE = 'OLDER_AGE'
    YOUNGER_AGE = 'YOUNGER_AGE'
    RELATIONSHIP_TYPE = 'RELATIONSHIP_TYPE'


class RelationshipType(StrEnum):
    TRANSITORY = 'TRANSITORY'
    INFORMAL = 'INFORMAL'
    MARITAL = 'MARITAL'
    COMMERCIAL = 'COMMERCIAL'
    COUNT = 'COUNT'


class CondomUsageParametersType(StrEnum):
    USE_DEFAULT = 'USE_DEFAULT'
    SPECIFY_USAGE = 'SPECIFY_USAGE'


class TargetDiseaseState(StrEnum):
    HIV_POSITIVE = 'HIV_Positive'
    HIV_NEGATIVE = 'HIV_Negative'
    TESTED_POSITIVE = 'Tested_Positive'
    TESTED_NEGATIVE = 'Tested_Negative'
    MALE_CIRCUMCISION_POSITIVE = 'Male_Circumcision_Positive'
    MALE_CIRCUMCISION_NEGATIVE = 'Male_Circumcision_Negative'
    HAS_INTERVENTION = 'Has_Intervention'
    NOT_HAVE_INTERVENTION = 'Not_Have_Intervention'


# __all_exports: A list of classes that are intended to be exported from this module.
__all_exports = [
    StrEnum,
    DistributionType,
    NodeSelectionType,
    VaccineType,
    SensitivityType,
    EventOrConfig,
    SettingType,
    PrioritizePartnersBy,
    RelationshipType,
    CondomUsageParametersType,
    TargetDiseaseState]


# The following loop sets the __module__ attribute of each class in __all_exports to the name of the current module.
# This is done to ensure that when these classes are imported from this module, their __module__ attribute correctly
# reflects their source module.

for _ in __all_exports:
    _.__module__ = __name__

# __all__: A list that defines the public interface of this module.
# This is essential to ensure that Sphinx builds documentation for these classes, including those that are imported
# from emodpy.
# It contains the names of all the classes that should be accessible when this module is imported using the syntax
# 'from module import *'.
# Here, it is set to the names of all classes in __all_exports.

__all__ = [_.__name__ for _ in __all_exports]



from typing import Dict

from emod_api.demographics.updateable import Updateable
from emodpy.utils.distributions import WeibullDistribution

from emodpy_hiv.demographics.condom_usage_parameters import CondomUsageParameters


class RelationshipParameters(Updateable):
    def __init__(self,
                 condom_usage: CondomUsageParameters = None,
                 coital_act_rate: float = 0,
                 duration: WeibullDistribution = None):
        """
        A RelationshipParameters object defines basic attributes such as relationship duration, what happens if one
        member of a relationship migrates, and condom usage. Only WeibullDistribution durations are accepted by EMOD.

        Args:
            condom_usage: (CondomUsageParameters) (optional) a CondomUsageParameters object defining condom usage over
                time
            coital_act_rate: (float) (optional) the number of coital acts per day in a relationship. Default is 0.
            duration: (WeibullDistribution) (optional) a weibull distribution used for determining relationship duration
                in days (default is a WeibullDistribution with lambda and kappa values of 0.1)
        """
        super().__init__()
        self.condom_usage = condom_usage if condom_usage is not None else CondomUsageParameters()
        self.coital_act_rate = coital_act_rate
        # TODO: do we need to allow alteration of migration actions??
        self.migration_actions = ["TERMINATE"]
        self.migration_actions_distribution = [1.0]

        default_duration = WeibullDistribution(weibull_lambda=0.1, weibull_kappa=0.1)
        self.duration = duration if duration is not None else default_duration

    def to_dict(self) -> Dict:
        duration_heterogeneity = self.duration.weibull_kappa ** -1
        parameters = {
            'Condom_Usage_Probability': self.condom_usage.to_dict(),
            'Coital_Act_Rate': self.coital_act_rate,
            'Duration_Weibull_Scale': self.duration.weibull_lambda,
            'Duration_Weibull_Heterogeneity': duration_heterogeneity,
            'Migration_Actions': self.migration_actions,
            'Migration_Actions_Distribution': self.migration_actions_distribution
        }
        return parameters

    @classmethod
    def from_dict(cls, d: Dict) -> '__class__':
        condom_usage = CondomUsageParameters.from_dict(d=d['Condom_Usage_Probability'])
        kappa = 0 if d['Duration_Weibull_Heterogeneity'] == 0 else d['Duration_Weibull_Heterogeneity'] ** -1
        duration = WeibullDistribution(weibull_lambda=d['Duration_Weibull_Scale'], weibull_kappa=kappa)
        return cls(condom_usage=condom_usage, coital_act_rate=d['Coital_Act_Rate'], duration=duration)



import os
import pandas as pd
from sklearn.linear_model import LinearRegression

import emodpy_hiv.plotting.xy_plot as xy_plot
import emodpy_hiv.plotting.helpers as helpers

COL_NAME_REL_ID             = "Rel_ID"           # noqa: E221
COL_NAME_START_TIME         = "Rel_start_time"   # noqa: E221
COL_NAME_SCHEDULED_END_TIME = "Rel_scheduled_end_time"
COL_NAME_REL_TYPE           = "Rel_type (0 = TRANSITORY; 1 = INFORMAL; 2 = MARITAL; 3 = COMMERCIAL)"  # noqa: E221
# "Is_rel_outside_PFA"
# "Original_node_ID"
# "Current_node_ID"
# "A_ID"
# "A_is_infected"
# "A_gender"
# "A_age"
COL_NAME_RISK_A             = "A_IP='Risk'"      # noqa: E221
# "A_previous_num_coital_acts"
# "A_total_num_active_rels"
# "A_num_active_TRANSITORY_rels"
# "A_num_active_INFORMAL_rels"
# "A_num_active_MARITAL_rels"
# "A_num_active_COMMERCIAL_rels"
# "A_num_lifetime_rels"
# "A_num_rels_last_6_mo"
# "A_extra_relational_bitmask"
# "A_is_circumcised"
# "A_has_STI_coinfection"
# "A_is_superspreader"
# "B_ID"
# "B_is_infected"
# "B_gender"
# "B_age"
COL_NAME_RISK_B             = "B_IP='Risk'"      # noqa: E221
# "B_previous_num_coital_acts"
# "B_total_num_active_rels"
# "B_num_active_TRANSITORY_rels"
# "B_num_active_INFORMAL_rels"
# "B_num_active_MARITAL_rels"
# "B_num_active_COMMERCIAL_rels"
# "B_num_lifetime_rels"
# "B_num_rels_last_6_mo"
# "B_extra_relational_bitmask"
# "B_is_circumcised"
# "B_has_STI_coinfection"
# "B_is_superspreader"
# "A_CD4_count"
# "B_CD4_count"
# "A_viral_load"
# "B_viral_load"
# "A_HIV_disease_stage"
# "B_HIV_disease_stage"
# "A_HIV_Tested_Positive"
# "B_HIV_Tested_Positive"
# "A_HIV_Received_Results"
# "B_HIV_Received_Results"


def extract_assortivity_risk(start_rel_filename: str,
                             relationship_type: int,
                             male_risk_value="LOW"):
    """
    For the given relationship type, extract the number of relationships that started
    during each time step for each risk value pair.  The male risk value is constant
    so it should return a dataframe with three columns.

    Args:
        start_rel_filename (str, required):
            The name and path of the RelationshipStart.csv file to be read.

        relationship_type (int, required):
            The type of relationship. Options: 0 (transitory), 1 (informal), 2 (marital), 3 (commercial).

        male_risk_value (str, optional):
            The risk value of the male in the relationship being plotted.  This will be
            either LOW, MEDIUM, or HIGH.  Capitalization matters.
            Default is LOW.

    Returns:
        Dataframe with three columns where each column is for a risk value pairing.
        Each row should represent a simulation time (in days) that had relationhips
        created of that time and risk value pairing.  There is no guarantee that
        relationships are created each time step.
    """
    df = pd.read_csv(start_rel_filename)

    if COL_NAME_REL_TYPE not in df.columns:
        raise ValueError(f"'{COL_NAME_REL_TYPE}' column does not exist in the file({start_rel_filename}).")

    if relationship_type not in df[COL_NAME_REL_TYPE].unique():
        raise ValueError(f"'{relationship_type}' does not appear as a relationship type in the file({start_rel_filename}).")

    df = df[ df[COL_NAME_REL_TYPE] == relationship_type ] # noqa: E201, E202

    results_df = pd.DataFrame()
    results_df.index = df[COL_NAME_START_TIME].unique()
    results_df[COL_NAME_START_TIME] = df[COL_NAME_START_TIME].unique()
    risk_values = ["LOW", "MEDIUM", "HIGH"]
    for rv_a in [male_risk_value]:
        df_risk = df[ df[COL_NAME_RISK_A] == rv_a ] # noqa: E201, E202
        for rv_b in risk_values:
            df_risk2 = df_risk[ df_risk[COL_NAME_RISK_B] == rv_b ] # noqa: E201, E202, E226
            tmp_df = pd.DataFrame()
            tmp_df[COL_NAME_START_TIME] = df_risk2[COL_NAME_START_TIME]
            tmp_df[COL_NAME_RISK_B] = df_risk2[COL_NAME_RISK_B]
            tmp_df = tmp_df.groupby(COL_NAME_START_TIME).count()
            if len(tmp_df[COL_NAME_RISK_B]) == 0:
                tmp_df = pd.DataFrame()
                tmp_df.index = df_risk[COL_NAME_START_TIME].unique()
                tmp_df[COL_NAME_START_TIME] = df_risk[COL_NAME_START_TIME].unique()
                tmp_df[COL_NAME_RISK_B] = 0
            results_df[rv_a + "-" + rv_b] = tmp_df[COL_NAME_RISK_B]
            results_df = results_df.fillna(0)

    del results_df[COL_NAME_START_TIME]
    results_df = results_df.fillna(0)

    return results_df


def plot_relationship_assortivity_risk(dir_or_filename: str,
                                       relationship_type: int,
                                       male_risk_value: str = "LOW",
                                       show_avg_per_run: bool = False,
                                       show_regression: bool = False,
                                       regression_dir: str = None,
                                       img_dir: str = None):
    """
    Create a plot showing the number of relationships of a given type
    that started during the timestep for a male with the give risk value
    versus females with the other possible values.  For example, if the
    male's risk value is HIGH, the plot will contain three curves:
    HIGH-LOW, HIGH-MEDIUM, and HIGH-HIGH.  They will all be for the
    given relationship type.  We only do three curves because the data
    can be quite noisy.

    The plot also has the option to show a least squares regression line
    for each risk value pair.  A CSV file can be saved with the regression
    data.  This can be used to compare with the plot_a_vs_b() function to
    compare the regression from two different sets of files.

    Args:
        dir_or_filename (str, required):
            The directory or filename containing the RelationshipStart.csv files.

        relationship_type (str, optional):
            The type of relationship. Options: 0 (transitory), 1 (informal), 2 (marital), 3 (commercial).

        male_risk_value (str, optional):
            The risk value of the male in the relationship being plotted.  This will be
            either LOW, MEDIUM, or HIGH.  Capitalization matters.
            Default is LOW.

        show_avg_per_run (bool, optional):
            If 'dir_or_filename' is a directory, this will calculate the average number of
            relationships started at each timestep for the different files in the directory.
            Default is False.

        show_regression (bool, optional):
            If true, a least squares regression line will be calculated and shown on the plot.
            There will be one line for each risk value pair.
            Default is False.

        regression_dir (str, optional):
            If 'show_regression' is true and this provides a path to a directory, then a CSV
            file will be saved with the data points of the displayed regression lines.  The
            name of the file will be the relationship type and the male's risk value.  For
            example, COMMERCIAL-HIGH.csv.
            Default is None.

        img_dir (str, optional):
            Directory to save the images. If None, the images will not be saved and a window will be opened.

    Returns:
        None - but image will be saved or window opened.
    """
    if not show_regression and regression_dir:
        raise ValueError("Regression directory is set but show_regression is False.\nYou need to show regression if you want to save it.")

    # -------------------------------------------------------------------
    # Get the list of RelationshipStart.csv files in the given directory.
    # -------------------------------------------------------------------
    dir_filenames = helpers.get_filenames(dir_or_filename=dir_or_filename,
                                          file_prefix="RelationshipStart",
                                          file_extension=".csv")

    # ----------------------------------------------------------------------------------
    # For the given relationship type, extract the number of relationships that started
    # at each time step for each risk value pair where the male risk value is constant.
    # This should result in three columns with counts of new relationships.
    # ----------------------------------------------------------------------------------
    combined_df = pd.DataFrame()
    for fn in dir_filenames:
        df = extract_assortivity_risk(start_rel_filename=fn,
                                      relationship_type=relationship_type,
                                      male_risk_value=male_risk_value)
        if len(combined_df.columns) == 0:
            combined_df.index = df.index
            for column_name in df.columns:
                name = column_name
                if not show_avg_per_run:
                    name = fn + "-" + name
                combined_df[name] = 0
        for column_name in df.columns:
            if show_avg_per_run:
                combined_df[column_name] = combined_df[column_name] + df[column_name]
            else:
                name = fn + "-" + column_name
                combined_df[name] = df[column_name]
        combined_df = combined_df.fillna(0)

    if show_avg_per_run:
        for column_name in combined_df.columns:
            combined_df[column_name] = combined_df[column_name] / len(dir_filenames)

    # ---------------------------------------------------------------------------------
    # For a given time step, convert the counts of new relationships to a fraction.
    # If we are show all the files, then we need to calculate the fractions for each
    # file.  If doing the average, then it will be the fractions based on the averages.
    # ---------------------------------------------------------------------------------
    col_name_prefixs = [""]
    if not show_avg_per_run:
        col_name_prefixs = []
        for fn in dir_filenames:
            col_name_prefixs.append(fn + "-")

    total_df = pd.DataFrame()
    total_df.index = combined_df.index

    for prefix in col_name_prefixs:
        total_label = "total-" + prefix
        total_df[total_label] = 0
        for column_name in combined_df.columns:
            if prefix in column_name:
                total_df[total_label] = total_df[total_label] + combined_df[column_name]
        for column_name in combined_df.columns:
            if prefix in column_name:
                combined_df[column_name] = combined_df[column_name] / total_df[total_label]
            combined_df[column_name] = combined_df[column_name].fillna(0)

    # ---------------------------------------------------------------------------------
    # If requested, determine a least squares regression line for each risk value pair.
    # ---------------------------------------------------------------------------------
    expected_df = None
    if show_regression:
        expected_df = pd.DataFrame()
        expected_df.index = combined_df.index
        x = combined_df.index.values
        x = x.reshape(len(x), 1)
        for col_name in combined_df.columns:
            y = combined_df[col_name].values      # Dependent variable
            y = y.reshape(len(y), 1)
            model = LinearRegression()
            model.fit(x, y)
            expected_df["Regression-" + col_name] = model.predict(x)

    # convert relationship type to a string
    rel_str = "TRANSITORY"
    if relationship_type == 1:
        rel_str = "INFORMAL"
    elif relationship_type == 2:
        rel_str = "MARITAL"
    elif relationship_type == 3:
        rel_str = "COMMERCIAL"

    # save the regression data to a file
    if regression_dir and show_regression:
        if not os.path.exists(regression_dir):
            os.makedirs(regression_dir)
        expected_df["Time"] = expected_df.index
        expected_df.to_csv(regression_dir + "/" + rel_str + "-" + male_risk_value + ".csv", index=False)
        del expected_df["Time"]

    # ------------------------------
    # Create the titles for the plot
    # ------------------------------
    title2 = ""
    for col in combined_df.columns:
        title2 = title2 + col + "=" + f'{combined_df[col].mean():0.3f}' + " "

    title = ""
    if show_avg_per_run:
        title = "Average Per Run - "
    title = title + f"Relationship Per Risk Assortivity - {rel_str} - Male Risk = {male_risk_value}"

    # -------------
    # plot the data
    # -------------
    xy_plot.xy_plot(img_dir=img_dir,
                    df=combined_df,
                    expected_df=expected_df,
                    title_1=title,
                    title_2=title2,
                    x_axis_name="Days",
                    y_axis_name="Fraction of Relationships",
                    show_legend=show_avg_per_run,
                    show_markers=show_avg_per_run,
                    fraction_of_total=False,
                    min_x=None, max_x=None, min_y=None, max_y=None,
                    x_axis_as_log_scale=False,
                    y_axis_as_log_scale=False)


def plot_relationship_assortivity_risk_all(dir_or_filename: str,
                                           regression_dir: str = None,
                                           img_dir: str = None):
    """
    Create a plot for each combination male risk value and relationship type.

    Args:
        dir_or_filename (str, required):
            The directory or filename containing the RelationshipStart.csv files or a specific file.

        regression_dir (str, optional):
            If provided, a CSV file will be created for each plot where the CSV file has one column
            for each risk value combination - three columns because the male's value is fixed.

        img_dir (str, optional):
            Directory to save the images.  If None, the images will not be saved and a window will be opened.
            Default is none - don't save image and open a window.

    Returns:
        None - but images will be saved or windows opened.
    """
    male_risk_values = ["LOW", "MEDIUM", "HIGH"]
    for rel_type in [0, 1, 2, 3]:
        for risk_value in male_risk_values:
            plot_relationship_assortivity_risk(dir_or_filename=dir_or_filename,
                                               relationship_type=rel_type,
                                               male_risk_value=risk_value,
                                               show_avg_per_run=True,
                                               show_regression=True,
                                               regression_dir=regression_dir,
                                               img_dir=img_dir)


if __name__ == '__main__':
    import argparse
    parser = argparse.ArgumentParser()

    parser.add_argument('dir_or_filename', type=str, nargs=1, help='A directory with RelationshipStart.csv files or a single file.')
    parser.add_argument('-o', '--output', default=None, help='If provided, a directory will be created and images saved to the folder.  If not provided, it opens windows.')
    parser.add_argument('-t', '--type_of_relationship', default='marital', help='Options: transitory, informal, marital, commercial')
    parser.add_argument('-r', '--risk_value_of_male', default='LOW', help='Options: LOW, MEDIUM, HIGH')
    # parser.add_argument('-n', '--node_id', type=int, default=None, help='Plot the data for a specific node.')
    parser.add_argument('-m', '--mean', help='Gives the average/mean of each run at that time point.', action='store_true')
    parser.add_argument('-e', '--estimated', help='Show the estimated regression line.', action='store_true')

    args = parser.parse_args()

    dir_or_filename = args.dir_or_filename[0]

    possible_risk_values = ["LOW", "MEDIUM", "HIGH"]
    if args.risk_value_of_male not in possible_risk_values:
        raise ValueError("Unknown risk value for male = " + args.risk_value_of_male)

    possible_relationship_types = ["transitory", "informal", "marital", "commercial"]
    if args.type_of_relationship not in possible_relationship_types:
        raise ValueError("Unknown relationship type = " + args.type_of_relationship)

    rel_type = 0
    if args.type_of_relationship == "transitory":
        rel_type = 0
    elif args.type_of_relationship == "informal":
        rel_type = 1
    elif args.type_of_relationship == "marital":
        rel_type = 2
    else:
        rel_type = 3

    plot_relationship_assortivity_risk(dir_or_filename=dir_or_filename,
                                       relationship_type=rel_type,
                                       male_risk_value=args.risk_value_of_male,
                                       show_avg_per_run=args.mean,
                                       show_regression=args.estimated,
                                       regression_dir=None,
                                       img_dir=args.output)



import os


def get_filenames(dir_or_filename: str,
                  file_prefix: str,
                  file_extension: str = None):
    """
    Get a list of filenames from a directory or a single file.
    If a directory is provided, it will list all files in that directory
    and in any subdirectory.  If a file is provided, it will check if it
    matches the prefix and extension.

    This should handle the situation where the user provides a directory
    that only contains a list of files, say from emodpy_workflow's download command.
    This should also plot the files in a directory that contains subdirectories
    created by the DownloadAnalyzer where each subdirectory contains the reports
    from a simulation.  Finally, this should also handle getting all of the files
    in a suite or experiment directory created when using SLURM or Container platform
    for the running of EMOD - the actual directories where the simulations are run.

    Args:
        dir_or_filename (string, required):
            Directory or filename to search.
        file_prefix (string, required)
            Prefix to filter files by.
        file_extension (string, optional):
            Extension to filter files by.
    Returns:
        List of filenames that match the criteria.
    """

    dir_filenames = []

    # Check if the input is a directory or a file
    if os.path.isdir(dir_or_filename):
        # If it's a directory, list all files in the directory
        for base_entry in os.listdir(dir_or_filename):
            entry = os.path.join(dir_or_filename, base_entry)

            if os.path.isdir(entry):
                # If it's a subdirectory, recursively get filenames from it
                sub_filenames = get_filenames(dir_or_filename=entry,
                                              file_prefix=file_prefix,
                                              file_extension=file_extension)
                dir_filenames.extend(sub_filenames)
                continue

            if not os.path.isfile(entry):
                continue
            if file_prefix and not base_entry.startswith(file_prefix):
                continue
            if file_extension and not base_entry.endswith(file_extension):
                continue
            dir_filenames.append(entry)

    elif os.path.isfile(dir_or_filename):
        if (file_prefix and dir_or_filename.startswith(file_prefix)) or \
           (file_extension and dir_or_filename.endswith(file_extension)):
            dir_filenames.append(dir_or_filename)
        elif file_prefix:
            raise ValueError(f"'{dir_or_filename}' does not start with the specified prefix '{file_prefix}'.")
        elif file_extension:
            raise ValueError(f"'{dir_or_filename}' does not end with the specified extension '{file_extension}'.")
        else:
            dir_filenames.append(dir_or_filename)

    else:
        raise ValueError(f"'{dir_or_filename}' is neither a valid directory nor a file.")

    dir_filenames = sorted(dir_filenames)
    return dir_filenames



"""
This module contains the classes and functions for creating demographics files
for HIV simulations. For more information on |EMOD_s| demographics files,
see :doc:`emod-hiv:emod/software-demographics`.
"""
import pandas as pd

from typing import List

from emod_api.demographics.fertility_distribution import FertilityDistribution
from emodpy.demographics.demographics import Demographics
from emod_api.demographics.properties_and_attributes import IndividualProperty
from emodpy.utils.distributions import UniformDistribution

from emodpy_hiv.demographics import DemographicsTemplates as hiv_dt
from emodpy_hiv.demographics.assortivity import Assortivity
from emodpy_hiv.demographics.hiv_node import HIVNode
from emodpy_hiv.demographics.society import Society
from emodpy_hiv.demographics.year_age_rate import YearAgeRate


class HIVDemographics(Demographics):
    def __init__(self, nodes: List[HIVNode], default_society_template: str = None):
        """
        This class is derived from :py:class:`emodpy:emodpy.demographics.demographics.Demographics` adding HIV-
        specific features and sets certain defaults for HIV in construction.

        Args:
            nodes: A list of (non-Default) HIVNode objects
            default_society_template: society template name for loading initial society information. Will apply
                to the Default node. Default society template is 'PFA-Southern-Africa'.

        Returns:
            an HIVDemographics object
         """
        # we need to generate the default node before calling super() because we need it to be an HIVNode, not Node
        society_dict = hiv_dt.get_society_dict(society_name=default_society_template)
        society = Society.from_dict(d=society_dict)
        default_node = HIVNode(name='Default', lat=0, lon=0, pop=0, forced_id=0, society=society)

        super().__init__(nodes=nodes, idref="EMOD-HIV world", default_node=default_node)

        # default individual properties for HIV
        # We are assuming Risk-based assortivity
        self.add_or_update_initial_risk_distribution(distribution=[0.6669671396606822, 0.3330328603393178, 0])

        # default individual attributes for HIV
        # Uniform age distribution by default
        initial_age_distribution = UniformDistribution(uniform_min=0, uniform_max=18250)
        self.set_age_distribution(distribution=initial_age_distribution)

        # Node attributes copied from Demographics.SetDefaultNodeAttributes (which uses self.raw, which cannot
        #   be used with HIVDemographics). These may or may not be needed by HIV.
        default_node.node_attributes.altitude = 0
        default_node.node_attributes.airport = 1
        default_node.node_attributes.region = 1
        default_node.node_attributes.seaport = 1

    @property
    def raw(self):
        raise AttributeError("raw is not a valid attribute for HIVDemographics objects")

    @raw.setter
    def raw(self, value):
        raise AttributeError("raw is not a valid attribute for HIVDemographics objects")

    def set_fertility_distribution(self,
                                   distribution: FertilityDistribution,
                                   node_ids: List[int] = None) -> None:
        """
        Sets a fertility distribution on the demographics object. Automatically handles any necessary config updates.
        Args:
            distribution: The distribution to set. Must be a FertilityDistribution object for a complex distribution.
            node_ids: The node id(s) to apply changes to. None or 0 means the default node.
        Returns:
            Nothing
        """
        from emod_api.demographics.implicit_functions import _set_fertility_age_year
        self._set_distribution(distribution=distribution,
                               use_case='fertility',
                               complex_distribution_implicits=[_set_fertility_age_year],
                               node_ids=node_ids)

    def set_concurrency_params_by_type_and_risk(self, relationship_type: str, risk_group: str,
                                                max_simul_rels_male: float = None, max_simul_rels_female: float = None,
                                                prob_xtra_rel_male: float = None, prob_xtra_rel_female: float = None,
                                                node_ids: List[int] = None) -> None:
        """
        Set concurrent relationship formation parameters for a given relationship type, risk group, and node(s).

        Only non-None values will be updated.

        Each agent updates their max values and their ability to have extra relationships when they change
        nodes/locations or change their Risk IP value.

        Notes on maximums:
            - A max relationships parameter can be set to 2+, but if probability of extra is zero, an agent will never
              have more than one relationship of the given type.
            - fractional max relationship parameters are probabilistically rounded up or down per agent. E.g., max
              relationships of 2.3 implies agents have a (70%, 30%) chance of getting a maximum of (2, 3), respectively.

        Note on probabilities:
            - These probabilities apply to agents currently with at least ONE relationship of the given type seeking
              an additional such relationship.

        Args:
            relationship_type: Relationship type to update: "COMMERCIAL", "MARITAL", "INFORMAL" or "TRANSITORY"
            risk_group: risk group to set concurrency parameters for. "HIGH", "MEDIUM", or "LOW"
            max_simul_rels_male: Sets the maximum simultaneous relationships of the given relationship_type for males
                in the given risk_group.
            max_simul_rels_female: Sets the maximum simultaneous relationships of the given relationship_type for
                females in the given risk_group.
            prob_xtra_rel_male: The probability of a male receiving a flag that allows him to seek additional
                relationships while currently in another relationship.
            prob_xtra_rel_female: The probability of a female receiving a flag that allows her to seek additional
                relationships while currently in another relationship.
            node_ids: the id(s) of node(s) to apply changes to. None or 0 refers to the Default node.

        Returns:
            None
        """
        for node in self.get_nodes_by_id(node_ids=node_ids).values():
            node.society.set_concurrency_parameters(relationship_type=relationship_type,
                                                    risk=risk_group,
                                                    max_simul_rels_male=max_simul_rels_male,
                                                    max_simul_rels_female=max_simul_rels_female,
                                                    prob_xtra_rel_male=prob_xtra_rel_male,
                                                    prob_xtra_rel_female=prob_xtra_rel_female)

    def set_pair_formation_parameters(self, relationship_type: str,
                                      formation_rate: float = None,
                                      assortivity_matrix: List[List[float]] = None,
                                      node_ids: List[int] = None) -> None:
        """
        Sets pair formation parameters for the specified relationship type and node(s).

        Only non-None values will be updated.

        Args:
            relationship_type: Relationship type to update: "COMMERCIAL", "MARITAL", "INFORMAL" or "TRANSITORY"
            assortivity_matrix: 3x3 row-major matrix of assortivity values, row represents male,
                column represents female. E.g. matrix[male_index][female_index]
                This matrix defines the tendency for a man of a particular Risk group to select a woman based on her
                Risk group (LOW/MEDIUM/HIGH male selecting LOW/MEDIUM/HIGH female -> 3x3 matrix).
            formation_rate: The number of new relationships per day for this relationship type that an individual will
                start. This can be changed during a simulation via a RelationshipFormationRateChanger intervention.
            node_ids: the id(s) of node(s) to apply changes to. None or 0 refers to the Default node.

        Returns:
            None
        """
        assortivity = None if assortivity_matrix is None else Assortivity(matrix=assortivity_matrix)
        for node in self.get_nodes_by_id(node_ids=node_ids).values():
            node.society.set_pair_formation_parameters(relationship_type=relationship_type,
                                                       formation_rate=formation_rate,
                                                       assortivity=assortivity)

    def set_relationship_parameters(self, relationship_type: str,
                                    coital_act_rate: float = None,
                                    condom_usage_min: float = None,
                                    condom_usage_mid: float = None,
                                    condom_usage_max: float = None,
                                    condom_usage_rate: float = None,
                                    duration_scale: float = None,
                                    duration_heterogeneity: float = None,
                                    node_ids: List[int] = None) -> None:
        """
        Sets relationship parameters for the specified relationship type and node(s).

        Only non-None values will be updated.

        Args:
            relationship_type: Relationship type to update: "COMMERCIAL", "MARITAL", "INFORMAL" or "TRANSITORY"
            coital_act_rate: Sets the per-day coital act rate for the specified relationship type. This can be changed
                during a simulation via a CoitalActRateChanger intervention.
            condom_usage_min: minimum condom usage probability (pre-inflection point). All four of these condom usage
                parameters can be changed during a simulation via a CondomUsageProbabilityChanger intervention.
            condom_usage_mid: inflection point in condom usage (a year)
            condom_usage_max: maximum condom usage probability (post-inflection point)
            condom_usage_rate: slope of condom usage at inflection point
            duration_scale: weibull distributed relationship duration value (Lambda)
            duration_heterogeneity: weibull distributed relationship heterogeneity value (1/Kappa)
            node_ids: the id(s) of node(s) to apply changes to. None or 0 refers to the Default node.

        Returns:
            None
        """
        for node in self.get_nodes_by_id(node_ids=node_ids).values():
            node.society.set_relationship_parameters(relationship_type=relationship_type,
                                                     coital_act_rate=coital_act_rate,
                                                     condom_usage_min=condom_usage_min,
                                                     condom_usage_mid=condom_usage_mid,
                                                     condom_usage_max=condom_usage_max,
                                                     condom_usage_rate=condom_usage_rate,
                                                     duration_scale=duration_scale,
                                                     duration_heterogeneity=duration_heterogeneity)

    def _add_or_update_individual_property_distribution(self, property_name: str,
                                                        values: List[str],
                                                        distribution: List[float],
                                                        node_ids: List[int] = None) -> None:
        self.AddIndividualPropertyAndHINT(Property=property_name, Values=values, InitialDistribution=distribution,
                                          node_ids=node_ids, overwrite_existing=True)

    # TODO: Should the following IP distribution setting methods accept lists of values, too? Or should we leave them
    #  hard-coded like this for "normal" usage, to be altered via _add_or_remove_... (above) if needed?
    #  https://github.com/InstituteforDiseaseModeling/emodpy-hiv/issues/206
    def add_or_update_initial_risk_distribution(self, distribution: List[float],
                                                node_ids: List[int] = None) -> None:
        """
        Adds the Risk individual property with specified initial distribution to the specified node(s).

        Args:
            distribution: a list of three floats that sum to 1 corresponding to distribution of Risk in this order:
                'LOW', 'MEDIUM', 'HIGH'
            node_ids: the id(s) of node(s) to apply changes to. None or 0 refers to the Default node.

        Returns:
            None
        """
        property = 'Risk'
        values = ['LOW', 'MEDIUM', 'HIGH']
        self._add_or_update_individual_property_distribution(property_name=property, values=values,
                                                             distribution=distribution, node_ids=node_ids)

    # TODO: change CascadeState here (and anywhere else) to InterventionStatus for consistency with EMOD terminology
    #  https://github.com/InstituteforDiseaseModeling/emodpy-hiv/issues/213
    def add_or_update_initial_cascade_state_distribution(self, distribution: List[float],
                                                         node_ids: List[int] = None) -> None:
        """
        Adds the CascadeState individual property with specified initial distribution to the specified node(s).

        Args:
            distribution: a list of fourteen floats that sum to 1 corresponding to distribution of CascadeState in this
                order:
                '', 'ARTStaging', 'ARTStagingDiagnosticTest', 'LinkingToART', 'LinkingToPreART', 'OnART', 'OnPreART',
                'HCTTestingLoop', 'HCTUptakeAtDebut', 'HCTUptakePostDebut', 'TestingOnANC', 'TestingOnChild6w',
                'TestingOnSymptomatic', 'LostForever'
            node_ids: the id(s) of node(s) to apply changes to. None or 0 refers to the Default node.

        Returns:
            None
        """
        property = 'CascadeState'
        values = ["", "ARTStaging", "ARTStagingDiagnosticTest", "LinkingToART", "LinkingToPreART", "OnART", "OnPreART",
                  "HCTTestingLoop", "HCTUptakeAtDebut", "HCTUptakePostDebut", "TestingOnANC", "TestingOnChild6w",
                  "TestingOnSymptomatic", "LostForever"]
        self._add_or_update_individual_property_distribution(property_name=property, values=values,
                                                             distribution=distribution, node_ids=node_ids)

    def add_or_update_initial_health_care_accessibility_distribution(self, distribution: List[float],
                                                                     node_ids: List[int] = None) -> None:
        """
        Adds the (health care) Accessibility individual property with specified initial distribution to the specified
            node(s).

        Args:
            distribution: a list of three floats that sum to 1 corresponding to distribution of Accessibility in this
                order:
                'Yes', 'No'
            node_ids: the id(s) of node(s) to apply changes to. None or 0 refers to the Default node.

        Returns:
            None
        """
        property = 'Accessibility'
        values = ['Yes', 'No']
        self._add_or_update_individual_property_distribution(property_name=property, values=values,
                                                             distribution=distribution, node_ids=node_ids)

    def AddIndividualPropertyAndHINT(self, Property: str, Values: List[str], InitialDistribution: List[float] = None,
                                     TransmissionMatrix: List[List[float]] = None, Transitions: List = None,
                                     node_ids: List[int] = None, overwrite_existing: bool = False) -> None:
        """
        Add Individual Properties, including an optional HINT configuration matrix.

        Individual properties act as 'labels' on model agents that can be used for identifying and targeting
        subpopulations in campaign elements and reports. E.g. model agents may be given a property ('Accessibility')
        that labels them as either having access to health care (value: 'Yes') or not (value: 'No').

        Property-based heterogeneous disease transmission (HINT) is available for generic, environmental, typhoid,
        airborne, or TBHIV simulations as other simulation types have parameters for modeling the heterogeneity of
        transmission. By default, transmission is assumed to occur homogeneously among the population within a node.

        Note: EMOD requires individual property key and values (Property and Values args) to be the same across all
            nodes. The individual distributions of individual properties (InitialDistribution) can vary acros nodes.

        Documentation of individual properties and HINT:
            https://docs.idmod.org/projects/emod-generic/en/latest/model-properties.html
            https://docs.idmod.org/projects/emod-generic/en/latest/model-hint.html

        Args:
            Property: a new individual property key to add (if property already exists an exception is raised
                unless overwrite_existing is True).
            Values: the valid values of the new property key
            InitialDistribution: The fractional initial distribution of each valid Values entry. Order must match
                Values argument.
            TransmissionMatrix: HINT transmission matrix.
            node_ids: The node ids to apply changes to. None or 0 means the 'Defaults' node.
            overwrite_existing: Determines if an error is thrown if the IP is found pre-existing at a specified node.
                False: throw exception. True: overwrite the existing property.
        Returns:
            None
        """
        # TODO: This does not play nicely with emod-api Demographics (used by malaria/other diseases). Need to remove
        #  this HIV-specific override method once demographics.raw is removed entirely from emod-api
        #  https://github.com/InstituteforDiseaseModeling/emod-api/issues/687
        nodes = self.get_nodes_by_id(node_ids=node_ids).values()
        for node in nodes:
            if not overwrite_existing and node.has_individual_property(property_key=Property):
                raise ValueError("Property Type '{0}' already present in IndividualProperties list".format(Property))

            # Check if Property is in whitelist. If not, auto-set Disable_IP_Whitelist
            # NOTE: Disable_IP_Whitelist is no longer in the malaria-ongoing branch of EMOD (used for malaria and HIV),
            # however, because other still-active branches of EMOD exist (that utilize Disable_IP_Whitelist), this logic
            # cannot be removed yet.
            ip_whitelist = ["Age_Bin", "Accessibility", "Geographic", "Place", "Risk", "QualityOfCare", "HasActiveTB",
                            "InterventionStatus"]
            if Property not in ip_whitelist:
                def update_config(config):
                    config.parameters["Disable_IP_Whitelist"] = 1
                    return config
                self.implicits.append(update_config)

            tm_dict = None if TransmissionMatrix is None else {"Route": "Contact", "Matrix": TransmissionMatrix}
            individual_property = IndividualProperty(property=Property,
                                                     values=Values,
                                                     initial_distribution=InitialDistribution,
                                                     transitions=Transitions,
                                                     transmission_matrix=tm_dict)
            node.individual_properties.add(individual_property=individual_property, overwrite=overwrite_existing)

        if TransmissionMatrix is not None:
            def update_config(config):
                config.parameters.Enable_Heterogeneous_Intranode_Transmission = 1
                return config
            self.implicits.append(update_config)

    @classmethod
    def from_population_dataframe(cls, df: pd.DataFrame, default_society_template: str = None) -> '__class__':
        """
        Loads per-node population information from a formatted pandas DataFrame object into HIVNode objects, returning
        an HIVDemographics object containing them.

        Expected DataFrame format:

        node_id,name,population
        1,Province1,1000
        2,Province2,2500
        ...


        Args:
            df: data for initializing the nodes of an
            default_society_template: society template name for loading initial society information. Will apply
                to the Default node. Default society template is 'PFA-Southern-Africa'.

        Returns:
            an HIVDemographics object
        """
        """
        TODO: add demographics file error checks
        - no duplicate node_ids
        - node_ids are integers 1+
        - no duplicate node names
        - populations are integers 0+
        - column checking: are exactly the required columns in the file?
        """
        # TODO: consider if this can be merged into a more general emod-api call
        nodes = []
        for index, row in df.iterrows():
            nodes.append(HIVNode(lat=0, lon=0, pop=int(row['population']),
                                 name=row['name'], forced_id=int(row['node_id'])))
        return cls(nodes=nodes, default_society_template=default_society_template)

    @classmethod
    def from_template_node(cls, lat: float = 0, lon: float = 0, pop: float = 1e6, name: str = 'node1',
                           forced_id: int = 1, default_society_template: str = None) -> '__class__':
        """
        Creates a single-node HIVDemographics object from the supplied parameters

        Args:
            lat: Latitude of the centroid of the node to create.
            lon: Longitude of the centroid of the node to create.
            pop: Human population of the node.
            name: The name of the node. This may be a characteristic of the node, such as "rural" or "urban", or an
                identifying feature or value.
            forced_id: The node ID for the single node.
            default_society_template: society template name for loading initial society information. Will apply
                to the Default node. Default society template is 'PFA-Southern-Africa'.

        Returns:
            An HIVDemographics object
        """
        new_nodes = [HIVNode(lat=lat, lon=lon, pop=pop, name=name, forced_id=forced_id)]
        return cls(nodes=new_nodes, default_society_template=default_society_template)

    # Disabling these functions (from the super-classes, to avoid having to test them as there is no
    # current workflow using these routes in initialization.

    @classmethod
    def from_pop_csv(cls):
        """
        This method of building demographics is not available for HIVDemographics
        """
        raise NotImplementedError('This method of building demographics is not available for HIVDemographics')

    @classmethod
    def from_params(cls):
        """
        This method of building demographics is not available for HIVDemographics
        """
        raise NotImplementedError('This method of building demographics is not available for HIVDemographics')

    @classmethod
    def from_year_age_rate_data(cls,
                                pop_df:               pd.DataFrame,  # noqa: E241 - node_id, name, population
                                age_distribution_yar: YearAgeRate,
                                fertility_yar:        YearAgeRate,   # noqa: E241
                                male_mortality_yar:   YearAgeRate,   # noqa: E241
                                female_mortality_yar: YearAgeRate,
                                society: Society = None):
        """
        Create an HIVDemographics object using YearAgeRate data for the initial age distribution,
        fertility, and mortality.

        Args:
            pop_df: A pandas dataframe with columns "node_id", "name", and "population" where:
                "node_id" is an unsigned integer ranging from 1 to  4,294,967,295
                "name" is a string that one can use when creating the demographics
                "population" is an unsigned integer ranging from 0 to  4,294,967,295
            age_distribution_yar: A YearAgeRate object containing data for the ages of the
                initial population
            fertility_yar: A YearAgeRate object containing the fertility rates to use during
                the simulation
            male_mortality_yar: A YearAgeRate object containing the male mortality rates to use
                during the simulation
            female_mortality_yar: A YearAgeRate object containing the fwmale mortality rates to use
                during the simulation
            society: A Society object defining how people form relationships and have coital acts.
                Defaults to None (to be set later).

        Returns:
            An HIVDemographics object
        """
        demog = cls.from_population_dataframe(df=pop_df)

        age_dist_list = age_distribution_yar.to_age_distributions()
        for node_id, age_dist in age_dist_list:
            demog.set_age_distribution(distribution=age_dist, node_ids=[node_id])

        fert_dist_list = fertility_yar.to_fertility_distributions()
        for node_id, fert_dist in fert_dist_list:
            demog.set_fertility_distribution(distribution=fert_dist, node_ids=[node_id])

        male_mort_dist_dict = male_mortality_yar.to_mortality_distributions()
        female_mort_dist_dict = female_mortality_yar.to_mortality_distributions()
        for node_id, male_mort_dist in male_mort_dist_dict.items():
            demog.set_mortality_distribution(distribution_male=male_mort_dist,
                                             distribution_female=female_mort_dist_dict[node_id],
                                             node_ids=[node_id])

        if society:
            demog.society = society

        return demog
